import torch

from torchio.data.subject import Subject
from torchio.transforms.preprocessing.intensity.normalization_transform import (
    NormalizationTransform,
    TypeMaskingMethod,
)


# adapted from https://github.com/fepegar/torchio/blob/master/torchio/transforms/preprocessing/intensity/z_normalization.py
class NonZeroNormalizeOnMaskedRegion(NormalizationTransform):
    """
    Subtract mean and divide by standard deviation.

    Args:
        masking_method: See
            :class:`~torchio.transforms.preprocessing.intensity.NormalizationTransform`.
        **kwargs: See :class:`~torchio.transforms.Transform` for additional keyword arguments.
    """

    def __init__(self, masking_method: TypeMaskingMethod = None, **kwargs):
        super().__init__(masking_method=masking_method, **kwargs)
        self.args_names = ("masking_method",)

    def apply_normalization(
        self, subject: Subject, image_name: str, mask: torch.Tensor
    ) -> None:
        image = subject[image_name]
        mask = image.data != 0
        standardized = self.znorm(image.data, mask)
        if standardized is None:
            message = (
                "Standard deviation is 0 for masked values"
                f' in image "{image_name}" ({image.path})'
            )
            raise RuntimeError(message)
        subject.get_images_dict(intensity_only=True)[image_name]["data"] = standardized

    @staticmethod
    def znorm(tensor: torch.Tensor, mask: torch.Tensor) -> torch.Tensor:
        tensor = tensor.clone().float()
        values = tensor.masked_select(mask)
        mean, std = values.mean(), values.std()
        if std == 0:
            return None
        tensor[mask] -= mean
        tensor[mask] /= std
        return tensor



from GANDLF.entrypoints.anonymizer import new_way as anonymizer_command
from GANDLF.entrypoints.run import new_way as run_command
from GANDLF.entrypoints.construct_csv import new_way as construct_csv_command
from GANDLF.entrypoints.collect_stats import new_way as collect_stats_command
from GANDLF.entrypoints.patch_miner import new_way as patch_miner_command
from GANDLF.entrypoints.preprocess import new_way as preprocess_command
from GANDLF.entrypoints.verify_install import new_way as verify_install_command
from GANDLF.entrypoints.config_generator import new_way as config_generator_command
from GANDLF.entrypoints.recover_config import new_way as recover_config_command
from GANDLF.entrypoints.deploy import new_way as deploy_command
from GANDLF.entrypoints.optimize_model import new_way as optimize_model_command
from GANDLF.entrypoints.generate_metrics import new_way as generate_metrics_command
from GANDLF.entrypoints.debug_info import new_way as debug_info_command
from GANDLF.entrypoints.split_csv import new_way as split_csv_command


cli_subcommands = {
    "anonymizer": anonymizer_command,
    "run": run_command,
    "construct-csv": construct_csv_command,
    "collect-stats": collect_stats_command,
    "patch-miner": patch_miner_command,
    "preprocess": preprocess_command,
    "config-generator": config_generator_command,
    "verify-install": verify_install_command,
    "recover-config": recover_config_command,
    "deploy": deploy_command,
    "optimize-model": optimize_model_command,
    "generate-metrics": generate_metrics_command,
    "debug-info": debug_info_command,
    "split-csv": split_csv_command,
}



from monai.optimizers import Novograd


def novograd_wrapper(parameters):
    return Novograd(
        parameters["model_parameters"],
        lr=parameters.get("learning_rate"),
        betas=parameters["optimizer"].get("betas", (0.9, 0.999)),
        eps=parameters["optimizer"].get("eps", 1e-8),
        weight_decay=parameters["optimizer"].get("weight_decay", 3e-05),
        amsgrad=parameters["optimizer"].get("amsgrad", False),
    )
